/***********************************************************************************[SolverTypes.h]
Copyright (c) 2003-2006, Niklas Een, Niklas Sorensson
Copyright (c) 2007-2010, Niklas Sorensson

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**************************************************************************************************/


#ifndef Minisat_SolverTypes_h
#define Minisat_SolverTypes_h

#define VSIDS 0
#define CHB 1
#define LRB 2

#ifndef BRANCHING_HEURISTIC
    #define BRANCHING_HEURISTIC 1 //TAK LRB
#endif

#ifndef LBD_BASED_CLAUSE_DELETION
    #define LBD_BASED_CLAUSE_DELETION true
#endif

#ifndef RAPID_DELETION
    #define RAPID_DELETION true
#endif

#ifndef ALMOST_CONFLICT
    #define ALMOST_CONFLICT false //TAK true
#endif
#if ALMOST_CONFLICT && BRANCHING_HEURISTIC != LRB
    #error ALMOST_CONFLICT requires BRANCHING_HEURISTIC == LRB
#endif

#ifndef ANTI_EXPLORATION
    #define ANTI_EXPLORATION false //TAK true
#endif
#if ANTI_EXPLORATION && BRANCHING_HEURISTIC != LRB
    #error ANTI_EXPLORATION requires BRANCHING_HEURISTIC == LRB
#endif
#define ALL_SAVE


#include <assert.h>

#include "mtl/IntTypes.h"
#include "mtl/Alg.h"
#include "mtl/Vec.h"
#include "mtl/Map.h"
#include "mtl/Alloc.h"
#include<bitset>
#include <map>
//#include <unordered_map>
#include <vector>

//#define USE_WATCHES_MAP
using namespace std;
namespace Minisat {
class Solver;
struct WEIGHT_AGGREGATE;
struct CARD_NODE;
struct CARD_ADDER_BASE;
struct TNode;
//=================================================================================================
// Variables, literals, lifted booleans, clauses:


// NOTE! Variables are just integers. No abstraction here. They should be chosen from 0..N,
// so that they can be used as array indices.

typedef int Var;
typedef int Value;

#define var_Undef (-1)

#if LBD_BASED_CLAUSE_DELETION
typedef int Act;
#else
typedef float Act;
#endif

struct Lit {
    int     x;

    // Use this as a constructor:
    friend Lit mkLit(Var var, bool sign = false);

    bool operator == (Lit p) const { return x == p.x; }
    bool operator != (Lit p) const { return x != p.x; }
    bool operator <  (Lit p) const { return x < p.x;  } // '<' makes p, ~p adjacent in the ordering.
};


inline  Lit  mkLit     (Var var, bool sign) { Lit p; p.x = var + var + (int)sign; return p; }
inline  Lit  operator ~(Lit p)              { Lit q; q.x = p.x ^ 1; return q; }
inline  Lit  operator ^(Lit p, bool b)      { Lit q; q.x = p.x ^ (unsigned int)b; return q; }
inline  bool sign      (Lit p)              { return p.x & 1; }
inline  int  var       (Lit p)              { return p.x >> 1; }

// Mapping Literals to and from compact integers suitable for array indexing:
inline  int  toInt     (Var v)              { return v; } 
inline  int  toInt     (Lit p)              { return p.x; } 
inline  Lit  toLit     (int i)              { Lit p; p.x = i; return p; } 

//const Lit lit_Undef = mkLit(var_Undef, false);  // }- Useful special constants.
//const Lit lit_Error = mkLit(var_Undef, true );  // }

const Lit lit_Undef = { -2 };  // }- Useful special constants.
const Lit lit_Error = { -1 };  // }


//=================================================================================================
// Lifted booleans:
//
// NOTE: this implementation is optimized for the case when comparisons between values are mostly
//       between one variable and one constant. Some care had to be taken to make sure that gcc 
//       does enough constant propagation to produce sensible code, and this appears to be somewhat
//       fragile unfortunately.

#define l_True  (lbool((uint8_t)0)) // gcc does not do constant propagation if these are real constants.
#define l_False (lbool((uint8_t)1))
#define l_Undef (lbool((uint8_t)2))
class timeout_exception {

};

class lbool {
    uint8_t value;

public:
    explicit lbool(uint8_t v) : value(v) { }

    lbool()       : value(0) { }
    explicit lbool(bool x) : value(!x) { }

    bool  operator == (lbool b) const { return ((b.value&2) & (value&2)) | (!(b.value&2)&(value == b.value)); }
    bool  operator != (lbool b) const { return !(*this == b); }
    lbool operator ^  (bool  b) const { return lbool((uint8_t)(value^(uint8_t)b)); }

    lbool operator && (lbool b) const { 
        uint8_t sel = (this->value << 1) | (b.value << 3);
        uint8_t v   = (0xF7F755F4 >> sel) & 3;
        return lbool(v); }

    lbool operator || (lbool b) const {
        uint8_t sel = (this->value << 1) | (b.value << 3);
        uint8_t v   = (0xFCFCF400 >> sel) & 3;
        return lbool(v); }

    friend int   toInt  (lbool l);
    friend lbool toLbool(int   v);
};
inline int   toInt  (lbool l) { return l.value; }
inline lbool toLbool(int   v) { return lbool((uint8_t)v);  }
inline unsigned get_mask_word(int context_mod_bit) {
	if (context_mod_bit <= 0) return 0;
	//	assert(context_mod_bit <= 32);
	//	if ((context_mod_bit % 32) == 0) return 0xffffffff;
	if (context_mod_bit >= 32) return 0xffffffff;//Sep.26.2016
	int m = 0x80000000;
	m >>= (31 - context_mod_bit);
	return ~m;
}
struct DeducedReason {
	Lit A;
	Lit B;

	DeducedReason() {
		clear();
	}
	void clear() {
		A = lit_Error;
		B = lit_Error;

	}

};



struct ConflictReason:public DeducedReason {

	Lit C;
	
	ConflictReason() {
		clear();

	}
	ConflictReason(DeducedReason dr) {
		A = dr.A;
		B = dr.B;

	}
	void clear() {
		A = lit_Error;
		B = lit_Error;
		C = lit_Error;
		
	}
	
};

struct Complex {
	unsigned aval;
	unsigned bval;
	Complex() {
		aval = 0xffffffff;
		bval = 0;

	}
	lbool get_lbool(unsigned mod_bit)const {
		unsigned bit_real = (aval >> mod_bit) & 0x01;
		unsigned bit_imag = (bval >> mod_bit) & 0x01;

		if (bit_imag) return l_Undef;
		else if (bit_real) return l_False;
		else return l_True;

	}
	void set_lbool(unsigned mod_bit, lbool v) {
		aval = (aval &  ~(1 << mod_bit)) | ((toInt(v) & 0x01) << mod_bit);
		bval = (bval &  ~(1 << mod_bit)) | (((toInt(v) >> 1) & 0x01) << mod_bit);
	}

	Complex get_and(lbool c) {


		if (c == l_False) {
			Complex C;
			return C;//SBIT@l_False

		}
		else if (c == l_True) {
			return *this;//


		}
		else {
			//l_False->l_False
			//l_True ->l_Undef
			//l_Undef ->l_Undef
			Complex C;
			C.aval = aval;
			C.bval = ~aval | bval;
			return C;
		}
	}
	//1BIT LEFT SHIFTBBit0́Al_Truepush
	Complex get_left_shift() {
		Complex C;
		C.aval = aval << 1;//
		C.bval = bval << 1;
		return C;
	}
	std::string get_s(int i) {
		if (get_lbool(i) == l_Undef) return "X";
		else if (get_lbool(i) == l_True) return "T";
		else if (get_lbool(i) == l_False) return "F";
		else assert(0);


	}

	//A&B !=BłȂƂɏol_FalseƂl_TrueɂȂĂƂ낪->Conflict
	//A@false܂undefBIT𗧂Ă
	unsigned make_false_or_undef_word() {
		return aval | bval;

	}
	//l_False BIT𗧂Ă
	unsigned make_false_word() {
		return aval & ~bval;
	}
	//l_TrueBIT𗧂Ă
	unsigned make_true_word() {
		unsigned temp = ~aval & ~bval;
		return ~aval & ~bval;
	}
	unsigned make_true_or_undef_word() {
		return ~aval | bval;
	}

	unsigned compare(Complex C) {
		unsigned compare = C.aval ^aval;//قȂӏ
		compare &= ~(C.bval | bval);//X́A
		return compare;
	}
	unsigned get_true_count() const {//l_True 0̐Ԃ
		unsigned a = ~aval & ~bval;
		return std::bitset<32>(a).count();
	}
	unsigned get_false_count() const {//l_True 0̐Ԃ
		unsigned a = aval & ~bval;
		return std::bitset<32>(a).count();
	}
	unsigned get_undef_count() const {//l_Undef̐Ԃ  Default Falseɐݒ肷̂ŁAFalseJEgƔ͈͊OJEgĂ܂

		return std::bitset<32>(bval).count();
	}
	unsigned get_true_count(unsigned MASK) const {//l_True 0̐Ԃ
		unsigned a = ~aval & ~bval &MASK;
		return std::bitset<32>(a).count();
	}
	unsigned get_undef_count(unsigned MASK) const {//l_Undef̐Ԃ  Default Falseɐݒ肷̂ŁAFalseJEgƔ͈͊OJEgĂ܂
		unsigned a = bval &MASK;
		return std::bitset<32>(a).count();
	}
	unsigned get_false_count(unsigned MASK) const {//l_Undef̐Ԃ  Default Falseɐݒ肷̂ŁAFalseJEgƔ͈͊OJEgĂ܂
		unsigned a = aval & ~bval&MASK;
		return std::bitset<32>(a).count();
	}
};
//w͈͓FalseĂƂtrue
template <int size, int  MASK>
inline bool has_false(Complex* C) {
	for (int i = 0;i < size - 1;i++) {
		if (C[i].make_false_word()) return true;//LSB WORD MSB猟
	}
	if (C[size - 1].make_false_word()&MASK) return true;//MSB猟

	return false;
}

//w͈͓TrueĂƂtrue
template <int size, int  MASK>
inline bool has_true(Complex* C) {
	if (C[size - 1].make_true_word()&MASK) return true;//MSB猟
	for (int i = size - 2;i >= 0;i--) {
		if (C[i].make_true_word()) return true;//LSB WORD MSB猟
	}
	return false;
}
//w͈͓XĂƂtrue
template <int size, int  MASK>
inline bool has_x(Complex* C) {
	if (C[size - 1].bval&MASK) return true;//MSB猟
	for (int i = size - 2;i >= 0;i--) {
		if (C[i].bval) return true;//LSB WORD MSB猟
	}
	return false;
}
/*
//TrueĂőPOS𓾂BŌWORD́AMASKŎw肷Bw͈͓TrueĂȂƂ́A-1
template <int size>
inline int get_tmax(Complex* C, unsigned MASK) {
	int msb = -1;
	for (int i = 0;i < size - 1;i++) {
		unsigned long Smsb;
		bool isNonzeroSomsb = _BitScanReverse(&Smsb, C[i].make_true_word());//LSB WORD MSB猟
		if (isNonzeroSomsb) {
			msb = i * 32 + Smsb;
		}
	}
	unsigned long Smsb;
	bool isNonzeroSomsb = _BitScanReverse(&Smsb, C[size - 1].make_true_word()&MASK);//MSB猟
	if (isNonzeroSomsb) {
		msb = (size - 1) * 32 + Smsb;
	}
	return msb;
}

//FalseĂŏPOS𓾂BŏWORD́AMASKŎw肷Bw͈͓FalseĂȂƂ́A-1
template <int size, int  MASK>
inline int get_fmin(Complex* C) {
	int lsb = -1;
	unsigned long Slsb;
	bool isNonzeroSolsb = _BitScanForward(&Slsb, C[size - 1].make_false_word()&MASK);//MSB Word LSB猟
	if (isNonzeroSolsb) {
		lsb = (size - 1) * 32 + Slsb;
	}
	for (int i = size - 2;i >= 0;i--) {

		unsigned long Slsb;
		bool isNonzeroSolsb = _BitScanForward(&Slsb, C[i].make_false_word());//MSB猟
		if (isNonzeroSolsb) {
			lsb = i * 32 + Slsb;
		}
	}
	return lsb;
}
//=================================================================================================

*/

// Clause -- a simple class for representing a clause:

class Clause;
struct Node;
struct NODE;
struct NODE_BASE;
struct NodeBottom;
struct SPECIAL_CLAUSE;

typedef RegionAllocator<uint32_t>::Ref CRef;

class Clause {
    struct {
		unsigned mark : 2;//2;sv
	//	unsigned assumption_var_used:1;
        unsigned learnt    : 1;
        unsigned has_extra : 1;
        unsigned reloced   : 1;
		unsigned type :	     3;//ClauseTYPE
		unsigned size : 23;//27;Clauseǉɔ
	}                            header;
    union { Lit lit; Act act; uint32_t abs; CRef rel; } data[0];

    friend class ClauseAllocator;
public:
	enum Type
	{
		NormalClause = 0, CARD_CLAUSE = 1, WEIGHT_CLAUSE = 2, MOSTONE = 3,CARDADDERBOTTOM=4,CARDADDER=5 };
    // NOTE: This constructor cannot be used directly (doesn't allocate enough memory).
    template<class V>
    Clause(const V& ps, bool use_extra, bool learnt,Type t=NormalClause) {
        header.mark      = 0;
//        header.assumption_var_used=0;
        header.learnt    = learnt;
        header.has_extra = use_extra;
        header.reloced   = 0;
		header.type = t;//^Cvǉ
        header.size      = ps.size();

        for (int i = 0; i < ps.size(); i++) 
            data[i].lit = ps[i];

        if (header.has_extra){
            if (header.learnt)
                data[header.size].act = 0; 
            else 
                calcAbstraction(); }
    }

public:
	Type get_type()const { return static_cast<Type>(header.type); }
	void set_type(Type t) { header.type = t; }
	bool         is_CARDADDER()      const { return get_type() == CARDADDER || get_type()==CARDADDERBOTTOM; }// = header.is_andor;}
	bool		 is_CARD()	  const { return get_type() == CARD_CLAUSE; }
	bool		 is_WEIGHT()  const { return get_type() == WEIGHT_CLAUSE; }
	bool		 is_MOSTONE() const { return get_type() == MOSTONE; }
    void calcAbstraction() {
        assert(header.has_extra);
        uint32_t abstraction = 0;
        for (int i = 0; i < size(); i++)
            abstraction |= 1 << (var(data[i].lit) & 31);
        data[header.size].abs = abstraction;  }


    int          size        ()      const   { return header.size; }
    void         shrink      (int i)         { assert(i <= size()); if (header.has_extra) data[header.size-i] = data[header.size]; header.size -= i; }
    void         pop         ()              { shrink(1); }
    bool         learnt      ()      const   { return header.learnt; }
    bool         has_extra   ()      const   { return header.has_extra; }
    uint32_t     mark        ()      const   { return header.mark; }
  
  
    void         mark        (uint32_t m)    { header.mark = m; }
    
    const Lit&   last        ()      const   { return data[header.size-1].lit; }

    bool         reloced     ()      const   { return header.reloced; }
    CRef         relocation  ()      const   { return data[0].rel; }
    void         relocate    (CRef c)        { header.reloced = 1; data[0].rel = c; }

    // NOTE: somewhat unsafe to change the clause in-place! Must manually call 'calcAbstraction' afterwards for
    //       subsumption operations to behave correctly.
    Lit&         operator [] (int i)         { return data[i].lit; }
    Lit          operator [] (int i) const   { return data[i].lit; }
    operator const Lit* (void) const         { return (Lit*)data; }

    Act&         activity    ()              { assert(header.has_extra); return data[header.size].act; }
    uint32_t     abstraction () const        { assert(header.has_extra); return data[header.size].abs; }

    Lit          subsumes    (const Clause& other) const;
    void         strengthen  (Lit p);
	Node*        get_node() {
		Node**p=(Node**)(&data[0]);
		return *p;
	}
	NODE_BASE*        get_NODE_BASE() {
		NODE_BASE**p = (NODE_BASE**)(&data[0]);
		return *p;
	}
	NodeBottom*        get_node_bottom() {
		NodeBottom**p= (NodeBottom**)(&data[0]);
		return *p;
	}
	SPECIAL_CLAUSE* get_special_clause() {
		SPECIAL_CLAUSE**p = (SPECIAL_CLAUSE**)(&data[0]);
		return *p;

	}
};


//=================================================================================================
// ClauseAllocator -- a simple class for allocating memory for clauses:


const CRef CRef_Undef = RegionAllocator<uint32_t>::Ref_Undef;
class ClauseAllocator : public RegionAllocator<uint32_t>
{
    static int clauseWord32Size(int size, bool has_extra){
        return (sizeof(Clause) + (sizeof(Lit) * (size + (int)has_extra))) / sizeof(uint32_t); }
 public:
    bool extra_clause_field;

    ClauseAllocator(uint32_t start_cap) : RegionAllocator<uint32_t>(start_cap), extra_clause_field(false){}
    ClauseAllocator() : extra_clause_field(false){}

    void moveTo(ClauseAllocator& to){
        to.extra_clause_field = extra_clause_field;
        RegionAllocator<uint32_t>::moveTo(to); }

    template<class Lits>
    CRef alloc(const Lits& ps, bool learnt = false,Clause::Type t = Clause::NormalClause)
    {
        assert(sizeof(Lit)      == sizeof(uint32_t));
        assert(sizeof(float)    == sizeof(uint32_t));
        bool use_extra = learnt | extra_clause_field;

        CRef cid = RegionAllocator<uint32_t>::alloc(clauseWord32Size(ps.size(), use_extra));
        new (lea(cid)) Clause(ps, use_extra, learnt,t);
		if (cid == 6933) {
			int m = 0;
		}
        return cid;
    }

    // Deref, Load Effective Address (LEA), Inverse of LEA (AEL):
    Clause&       operator[](Ref r)       { return (Clause&)RegionAllocator<uint32_t>::operator[](r); }
    const Clause& operator[](Ref r) const { return (Clause&)RegionAllocator<uint32_t>::operator[](r); }
    Clause*       lea       (Ref r)       { return (Clause*)RegionAllocator<uint32_t>::lea(r); }
    const Clause* lea       (Ref r) const { return (Clause*)RegionAllocator<uint32_t>::lea(r); }
    Ref           ael       (const Clause* t){ return RegionAllocator<uint32_t>::ael((uint32_t*)t); }

    void free(CRef cid)
    {
        Clause& c = operator[](cid);
        RegionAllocator<uint32_t>::free(clauseWord32Size(c.size(), c.has_extra()));
    }

    void reloc(CRef& cr, ClauseAllocator& to)
    {
        Clause& c = operator[](cr);
        
        if (c.reloced()) { cr = c.relocation(); return; }
        
        cr = to.alloc(c, c.learnt(),c.get_type());//TAK
        c.relocate(cr);
        
        // Copy extra data-fields: 
        // (This could be cleaned-up. Generalize Clause-constructor to be applicable here instead?)
        to[cr].mark(c.mark());
        
        if (to[cr].learnt())         to[cr].activity() = c.activity();
        else if (to[cr].has_extra()) to[cr].calcAbstraction();
    }
};

struct VarValue {
	Var v;
	unsigned type : 1;
	int value : 31;
	
	
	VarValue(Value val_, Var v_,bool tmax) {
		v = v_;
		value = val_;
		type = tmax;
	}
	int get_value() { return value; }
	Var get_var() { return v; }
	bool is_tmax() { return type; }
	bool is_fmin() { return !type; }
	bool get_type() { return type; }

	bool value_type_equal(VarValue V) {
		if (get_type() == V.get_type() && get_value() == V.get_value()) return true;
		else return false;
	}

};
struct AllState{
	CRef cref;
	int level;
	Value val;
	DeducedReason dr;
	AllState() {
		cref = CRef_Undef;
		level = 0;
		val = -1;


	}
	AllState(CRef cref_, int level_, Value val_, DeducedReason dr_) {
		cref = cref_;
		level = level_;
		val = val_;
		dr = dr_;
	}
};


class Solver;

struct NODE;

struct TF{
	short tmax;
	short fmin;

		
	
};	
struct port_idx{
	unsigned var_start;//4byte
	TF  tf;//4byte	
	
};	

union ptr_index {
	NODE_BASE* ptr;
	port_idx port;
};
struct vec_lit_level {
	vec_lit_level() {
		level = 0;

	}

	vec<Lit> lits;
	int level;


};

struct Tree {
	Tree*parent;
	Tree*childA;
	Tree*childB;

	int SAi;
	int SBi;
	
	vec<Lit> A_inputs, B_inputs;
	int layer;
	int K;
	int noutputs;

	Tree(Tree*parent_,  vec<Lit> &A_inputs_, vec<Lit>&B_inputs_,int K_, int	layer_) {
		parent = parent_;
		childA = 0;
		childB = 0;
		SAi=-1;
		SBi=-1;
		A_inputs_.copyTo(A_inputs);// = A_inputs_;
		B_inputs_.copyTo(B_inputs);
		K=K_;
		layer = layer_;
		noutputs = 0;
	}
	~Tree() {
		if (childA) delete childA;
		if (childB) delete childB;


	}
	void add_tree_map(map<int, map<int, Tree*> > &tree_map) {
		map<int, map<int, Tree*> >::iterator ix = tree_map.find(layer);
		if (ix == tree_map.end()) {
			map<int, Tree*> map;
			map[var(B_inputs[0])] = this;//B́A_~[ɂ邱Ƃ͂Ȃ

			tree_map[layer] = map;

		}
		else {
			tree_map[layer][var(B_inputs[0])] = this;
		}
		if (childA) childA->add_tree_map(tree_map);
		if (childB) childB->add_tree_map(tree_map);


	}
	void calc_noutputs() {
		if (SAi !=-1 && A_inputs.size()==1) {//A Port́A_~[@B@Port݂̂^̓ 
			noutputs = B_inputs.size();
			if (B_inputs.size() > K) {
				noutputs = K + 1;
			}
		}else {
			noutputs = A_inputs.size() + B_inputs.size();
			if (noutputs > K) {
				noutputs = K + 1;
			}

		}

	}

};
struct SPECIAL_CLAUSE
{
	virtual CRef prop(Solver&S, Lit L,CRef cr) { assert(0);return false; }
	virtual void set_tmax_recal_required() {  }
	virtual void set_fmin_recal_required() {  }
	virtual Clause::Type get_type() {
		assert(0); return Clause::NormalClause;
	}
	virtual	void attach_port(Solver&S,CRef cr){}
	virtual void Recal(Solver&S){}
	virtual void set_reason(Solver&S, vec<Lit>& lits,Lit L=lit_Undef){}

	CRef my_cref;
};




struct Watcher {
	CRef cref;
	Lit  blocker;
#ifdef USE_COPY_CONSTRUCTOR_VEC
	Watcher() {
		cref = CRef_Undef;
		blocker = lit_Undef;

	}
#endif
	Watcher(CRef cr, Lit p) : cref(cr), blocker(p) {}
	bool operator==(const Watcher& w) const { return cref == w.cref; }
	bool operator!=(const Watcher& w) const { return cref != w.cref; }
};

struct vec_watcher_bool {
	vec<Watcher> vec_watcher;
	bool dirty;
	vec_watcher_bool() {
		dirty = false;
	}
#ifdef	USE_COPY_CONSTRUCTOR_VEC
	vec_watcher_bool & operator=(const vec_watcher_bool vsb) {
		dirty = vsb.dirty;
		vec_watcher = vsb.vec_watcher;
		return *this;
	}
#endif
}
;

//=================================================================================================
// OccLists -- a class for maintaining occurence lists with lazy deletion:

template<class Idx, class Vec, class Deleted>
class OccLists
{
    vec<Vec>  occs;
    vec<char> dirty;
    vec<Idx>  dirties;
    Deleted   deleted;

 public:
    OccLists(const Deleted& d) : deleted(d) {}
    
    void  init      (const Idx& idx){ occs.growTo(toInt(idx)+1); dirty.growTo(toInt(idx)+1, 0); }
    // Vec&  operator[](const Idx& idx){ return occs[toInt(idx)]; }
    Vec&  operator[](const Idx& idx){ return occs[toInt(idx)]; }
    Vec&  lookup    (const Idx& idx){ if (dirty[toInt(idx)]) clean(idx); return occs[toInt(idx)]; }

    void  cleanAll  ();
    void  clean     (const Idx& idx);
    void  smudge    (const Idx& idx){
        if (dirty[toInt(idx)] == 0){
            dirty[toInt(idx)] = 1;
            dirties.push(idx);
        }
    }

    void  clear(bool free = true){
        occs   .clear(free);
        dirty  .clear(free);
        dirties.clear(free);
    }
};


template<class Idx, class Vec, class Deleted>
void OccLists<Idx,Vec,Deleted>::cleanAll()
{
    for (int i = 0; i < dirties.size(); i++)
        // Dirties may contain duplicates so check here if a variable is already cleaned:
        if (dirty[toInt(dirties[i])])
            clean(dirties[i]);
    dirties.clear();
}


template<class Idx, class Vec, class Deleted>
void OccLists<Idx,Vec,Deleted>::clean(const Idx& idx)
{
    Vec& vec = occs[toInt(idx)];
    int  i, j;
    for (i = j = 0; i < vec.size(); i++)
        if (!deleted(vec[i]))
            vec[j++] = vec[i];
    vec.shrink(i - j);
    dirty[toInt(idx)] = 0;
}
#ifdef USE_COPY_CONSTRUCTOR_VEC
template<class Idx, class Vec, class Deleted>
class OccMap
{
	//vec<Vec>  occs;
	//vec<char> dirty;
	std::map<int, Vec>  occs;
//	std::unordered_map<int, Vec> occs;
	vec<Idx>  dirties;
	Deleted   deleted;

public:
	OccMap(const Deleted& d) : deleted(d) {}

	void  init(const Idx& idx) {
		//occs.growTo(toInt(idx) + 1); dirty.growTo(toInt(idx) + 1, 0);
			
		vec_watcher_bool v;
		occs[toInt(idx)] = v;
	}
	// Vec&  operator[](const Idx& idx){ return occs[toInt(idx)]; }
	vec<Watcher>&  operator[](const Idx& idx) { return occs[toInt(idx)].vec_watcher; }


	void  cleanAll();
	void  clean(const Idx& idx);
	void  smudge(const Idx& idx) {
		if (!occs[toInt(idx)].dirty) {
			occs[toInt(idx)].dirty = true;
			dirties.push(idx);
		}
/*		if (dirty[toInt(idx)] == 0) {
			dirty[toInt(idx)] = 1;
			dirties.push(idx);
		}*/
	}

	void  clear(bool free = true) {
		occs.clear();
//		dirty.clear(free);
		dirties.clear(free);
	}
};


template<class Idx, class Vec, class Deleted>
void OccMap<Idx, Vec, Deleted>::cleanAll()
{
	for (int i = 0; i < dirties.size(); i++)
		// Dirties may contain duplicates so check here if a variable is already cleaned:
		if ( occs[toInt(dirties[i])].dirty)
			clean(dirties[i]);
	dirties.clear();

/*	for (int i = 0; i < dirties.size(); i++)
		// Dirties may contain duplicates so check here if a variable is already cleaned:
		if (dirty[toInt(dirties[i])])
			clean(dirties[i]);
	dirties.clear();*/
}


template<class Idx, class Vec, class Deleted>
void OccMap<Idx, Vec, Deleted>::clean(const Idx& idx)
{
	vec<Watcher>& V = occs[toInt(idx)].vec_watcher;
	int  i, j;
	for (i = j = 0; i < V.size(); i++)
		if (!deleted(V[i]))
			V[j++] = V[i];
	V.shrink(i - j);
	occs[toInt(idx)].dirty =false;


/*	Vec& vec = occs[toInt(idx)];
	int  i, j;
	for (i = j = 0; i < vec.size(); i++)
		if (!deleted(vec[i]))
			vec[j++] = vec[i];
	vec.shrink(i - j);
	dirty[toInt(idx)] = 0;*/
}
#endif
//=================================================================================================
// CMap -- a class for mapping clauses to values:


template<class T>
class CMap
{
    struct CRefHash {
        uint32_t operator()(CRef cr) const { return (uint32_t)cr; } };

    typedef Map<CRef, T, CRefHash> HashTable;
    HashTable map;
        
 public:
    // Size-operations:
    void     clear       ()                           { map.clear(); }
    int      size        ()                const      { return map.elems(); }

    
    // Insert/Remove/Test mapping:
    void     insert      (CRef cr, const T& t){ map.insert(cr, t); }
    void     growTo      (CRef cr, const T& t){ map.insert(cr, t); } // NOTE: for compatibility
    void     remove      (CRef cr)            { map.remove(cr); }
    bool     has         (CRef cr, T& t)      { return map.peek(cr, t); }

    // Vector interface (the clause 'c' must already exist):
    const T& operator [] (CRef cr) const      { return map[cr]; }
    T&       operator [] (CRef cr)            { return map[cr]; }

    // Iteration (not transparent at all at the moment):
    int  bucket_count() const { return map.bucket_count(); }
    const vec<typename HashTable::Pair>& bucket(int i) const { return map.bucket(i); }

    // Move contents to other map:
    void moveTo(CMap& other){ map.moveTo(other.map); }

    // TMP debug:
    void debug(){
        printf(" --- size = %d, bucket_count = %d\n", size(), map.bucket_count()); }
};


/*_________________________________________________________________________________________________
|
|  subsumes : (other : const Clause&)  ->  Lit
|  
|  Description:
|       Checks if clause subsumes 'other', and at the same time, if it can be used to simplify 'other'
|       by subsumption resolution.
|  
|    Result:
|       lit_Error  - No subsumption or simplification
|       lit_Undef  - Clause subsumes 'other'
|       p          - The literal p can be deleted from 'other'
|________________________________________________________________________________________________@*/
inline Lit Clause::subsumes(const Clause& other) const
{
    //if (other.size() < size() || (extra.abst & ~other.extra.abst) != 0)
    //if (other.size() < size() || (!learnt() && !other.learnt() && (extra.abst & ~other.extra.abst) != 0))
    assert(!header.learnt);   assert(!other.header.learnt);
    assert(header.has_extra); assert(other.header.has_extra);
    if (other.header.size < header.size || (data[header.size].abs & ~other.data[other.header.size].abs) != 0)
        return lit_Error;

    Lit        ret = lit_Undef;
    const Lit* c   = (const Lit*)(*this);
    const Lit* d   = (const Lit*)other;

    for (unsigned i = 0; i < header.size; i++) {
        // search for c[i] or ~c[i]
        for (unsigned j = 0; j < other.header.size; j++)
            if (c[i] == d[j])
                goto ok;
            else if (ret == lit_Undef && c[i] == ~d[j]){
                ret = c[i];
                goto ok;
            }

        // did not find it
        return lit_Error;
    ok:;
    }

    return ret;
}

inline void Clause::strengthen(Lit p)
{
    remove(*this, p);
    calcAbstraction();
}

//=================================================================================================
}

#endif
